# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

XCOLL_ELEMENTS_INIT_DEFAULTS = {
    # 'FlukaCollimator': {
    #     'length': 1., 'material': 'mogr'
    # },
    # 'FlukaCrystal': {
    #     'length': 1., 'material': 'si', 'bending_radius': 10., 'side': '+'
    # },
}
